# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/Norm_indices.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//Norm_indices
"""

from os import linesep
from sys import getsizeof
from ._check import check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from .Normalization import Normalization

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.Norm_indices.normalize import normalize
except ImportError as error:
    normalize = error


from ._check import InitUnKnowClassError


class Norm_indices(Normalization):
    """Return indices instead of axis values"""

    VERSION = 1

    # cf Methods.Norm_indices.normalize
    if isinstance(normalize, ImportError):
        normalize = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use Norm_indices method normalize: " + str(normalize)
                )
            )
        )
    else:
        normalize = normalize
    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(self, unit="SI", init_dict=None, init_str=None):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
        # Set the properties (value check and convertion are done in setter)
        # Call Normalization init
        super(Norm_indices, self).__init__(unit=unit)
        # The class is frozen (in Normalization init), for now it's impossible to
        # add new properties

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        Norm_indices_str = ""
        # Get the properties inherited from Normalization
        Norm_indices_str += super(Norm_indices, self).__str__()
        return Norm_indices_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False

        # Check the properties inherited from Normalization
        if not super(Norm_indices, self).__eq__(other):
            return False
        return True

    def compare(self, other, name="self", ignore_list=None):
        """Compare two objects and return list of differences"""

        if ignore_list is None:
            ignore_list = list()
        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()

        # Check the properties inherited from Normalization
        diff_list.extend(super(Norm_indices, self).compare(other, name=name))
        # Filter ignore differences
        diff_list = list(filter(lambda x: x not in ignore_list, diff_list))
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object

        # Get size of the properties inherited from Normalization
        S += super(Norm_indices, self).__sizeof__()
        return S

    def as_dict(self, type_handle_ndarray=0, keep_function=False, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        type_handle_ndarray: int
            How to handle ndarray (0: tolist, 1: copy, 2: nothing)
        keep_function : bool
            True to keep the function object, else return str
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        # Get the properties inherited from Normalization
        Norm_indices_dict = super(Norm_indices, self).as_dict(
            type_handle_ndarray=type_handle_ndarray,
            keep_function=keep_function,
            **kwargs
        )
        # The class name is added to the dict for deserialisation purpose
        # Overwrite the mother class name
        Norm_indices_dict["__class__"] = "Norm_indices"
        return Norm_indices_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        # Set to None the properties inherited from Normalization
        super(Norm_indices, self)._set_None()



from numpy import take
from SciDataTool.Functions.symmetries import rebuild_symmetries


def get_field(self, axes_list):
    """Returns the values of the field (with symmetries and sums).
    Parameters
    ----------
    self: Data
        a Data object
    axes_list: list
        a list of RequestedAxis objects
    Returns
    -------
    values: ndarray
        values of the field
    """

    values = self.values
    for axis_requested in axes_list:
        # Rebuild symmetries when needed
        axis_symmetries = self.axes[axis_requested.index].symmetries
        if axis_requested.is_pattern and (
            axis_requested.transform == "fft"
            or axis_requested.extension
            in ["sum", "rss", "mean", "rms", "integrate", "derivate", "antiderivate"]
            and axis_requested.is_pattern
        ):
            # DataPattern case where all values are requested
            values = take(values, axis_requested.rebuild_indices, axis_requested.index)
        elif axis_requested.transform == "fft" and "antiperiod" in axis_symmetries:
            # FFT case if axis is anti-periodic
            nper = axis_symmetries["antiperiod"]
            axis_symmetries["antiperiod"] = 2
            values = rebuild_symmetries(values, axis_requested.index, axis_symmetries)
            axis_symmetries["antiperiod"] = nper
        elif (
            axis_requested.indices is not None
            and max(axis_requested.indices) > values.shape[axis_requested.index]
        ):
            # Slicing case where requested indices are among other periods
            values = rebuild_symmetries(values, axis_requested.index, axis_symmetries)
            self.axes[axis_requested.index].symmetries = dict()

    return values



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/Normalization.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//Normalization
"""

from os import linesep
from sys import getsizeof
from ._check import check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from ._frozen import FrozenClass

from ._check import InitUnKnowClassError


class Normalization(FrozenClass):
    """Abstract class for normalizations"""

    VERSION = 1

    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(self, unit="SI", init_dict=None, init_str=None):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
        # Set the properties (value check and convertion are done in setter)
        self.parent = None
        self.unit = unit

        # The class is frozen, for now it's impossible to add new properties
        self._freeze()

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        Normalization_str = ""
        if self.parent is None:
            Normalization_str += "parent = None " + linesep
        else:
            Normalization_str += (
                "parent = " + str(type(self.parent)) + " object" + linesep
            )
        Normalization_str += 'unit = "' + str(self.unit) + '"' + linesep
        return Normalization_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False
        if other.unit != self.unit:
            return False
        return True

    def compare(self, other, name="self", ignore_list=None):
        """Compare two objects and return list of differences"""

        if ignore_list is None:
            ignore_list = list()
        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()
        if other._unit != self._unit:
            diff_list.append(name + ".unit")
        # Filter ignore differences
        diff_list = list(filter(lambda x: x not in ignore_list, diff_list))
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object
        S += getsizeof(self.unit)
        return S

    def as_dict(self, type_handle_ndarray=0, keep_function=False, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        type_handle_ndarray: int
            How to handle ndarray (0: tolist, 1: copy, 2: nothing)
        keep_function : bool
            True to keep the function object, else return str
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        Normalization_dict = dict()
        Normalization_dict["unit"] = self.unit
        # The class name is added to the dict for deserialisation purpose
        Normalization_dict["__class__"] = "Normalization"
        return Normalization_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        self.unit = None

    def _get_unit(self):
        """getter of unit"""
        return self._unit

    def _set_unit(self, value):
        """setter of unit"""
        check_var("unit", value, "str")
        self._unit = value

    unit = property(
        fget=_get_unit,
        fset=_set_unit,
        doc=u"""Normalization must be applied on axis values in this unit

        :Type: str
        """,
    )



import numpy as np

from SciDataTool.Functions.derivation_integration import integrate


def my_sum(values, index, Nper, is_aper):
    """Returns the arithmetic sum of values along given axis

    Parameters
    ----------
    values: ndarray
        array to derivate
    index: int
        index of axis along which to derivate
    Nper: int
        number of periods to replicate
    is_aper: bool
        True if values is anti-periodic along axis

    Returns
    -------
    values: ndarray
        arithmetic sum of values
    """

    if is_aper:
        # Sum of anti-periodic signal yields zero
        shape = list(values.shape)
        shape0 = [s for ii, s in enumerate(shape) if ii != index]
        values = np.zeros(shape0, dtype=values.dtype)
    else:
        # Take sum value multiplied by periodicity
        if Nper is None:
            # Set Nper to 1 in case of non-periodic axis
            Nper = 1
        values = Nper * np.sum(values, axis=index, keepdims=True)

    return values


def my_mean(values, ax_val, index, Nper, is_aper, is_phys):
    """Returns the mean (arithmetic or integral) of values along given axis

    Parameters
    ----------
    values: ndarray
        array to derivate
    ax_val: ndarray
        axis values
    index: int
        index of axis along which to derivate
    Nper: int
        number of periods to replicate
    is_aper: bool
        True if values is anti-periodic along axis
    is_phys: bool
        True if physical quantity (time/angle/z)

    Returns
    -------
    values: ndarray
        mean of values
    """

    if is_phys:
        # Integrate values and take mean value by dividing by integration interval in integrate()
        values = integrate(values, ax_val, index, Nper, is_aper, is_phys, is_mean=True)
    else:
        if is_aper:
            # Average of anti-periodic signal yields zero
            shape = list(values.shape)
            shape0 = [s for ii, s in enumerate(shape) if ii != index]
            values = np.zeros(shape0, dtype=values.dtype)
        else:
            # Take mean value multiplied by periodicity
            if Nper is None:
                # Set Nper to 1 in case of non-periodic axis
                Nper = 1
            values = Nper * np.mean(values, axis=index)

    return values


def root_mean_square(values, ax_val, index, Nper, is_aper, is_phys):
    """Returns the root mean square (arithmetic or integral) of values along given axis

    Parameters
    ----------
    values: ndarray
        array to derivate
    ax_val: ndarray
        axis values
    index: int
        index of axis along which to derivate
    Nper: int
        number of periods to replicate
    is_aper: bool
        True if values is anti-periodic along axis
    is_phys: bool
        True if physical quantity (time/angle/z)

    Returns
    -------
    values: ndarray
        root mean square of values
    """

    if is_aper and Nper is not None:
        # Remove anti-periodicity since values is squared
        is_aper = False

    return np.sqrt(my_mean(values ** 2, ax_val, index, Nper, is_aper, is_phys))


def root_sum_square(values, ax_val, index, Nper, is_aper, is_phys):
    """Returns the root sum square (arithmetic or integral) of values along given axis

    Parameters
    ----------
    values: ndarray
        array to derivate
    ax_val: ndarray
        axis values
    index: int
        index of axis along which to derivate
    Nper: int
        number of periods to replicate
    is_aper: bool
        True if values is anti-periodic along axis
    is_phys: bool
        True if physical quantity (time/angle/z)

    Returns
    -------
    values: ndarray
        root sum square of values
    """

    if is_aper and Nper is not None:
        # Remove anti-periodicity since values is squared
        is_aper = False

    if is_phys:
        values = integrate(values ** 2, ax_val, index, Nper, is_aper, is_phys)
    else:
        values = my_sum(values ** 2, index, Nper, is_aper)

    return np.sqrt(values)



def get_vectorfield_along(self, *args, unit="SI", is_norm=False, axis_data=[]):
    """Returns the sliced or interpolated version of the data, using conversions and symmetries if needed.
    Parameters
    ----------
    self : VectorField
        a VectorField object
    Returns
    -------
    a VectorField object
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.VectorField", fromlist=["VectorField"])
    VectorField = getattr(module, "VectorField")

    comp_dict = dict()

    for comp in self.components:  # Call get_data_along on each component
        comp_dict[comp] = self.components[comp].get_data_along(
            *args, unit=unit, is_norm=is_norm, axis_data=axis_data
        )

    return VectorField(name=self.name, symbol=self.symbol, components=comp_dict)



def normalize(
    self,
    values,
):
    """Normalizes axis values
    Parameters
    ----------
    self: Norm_func
        a Norm_func object
    values: ndarray
        axis values
    Returns
    -------
    Vector of axis values
    """
    return self.function(values)



from numpy import abs as np_abs

from SciDataTool.Functions.derivation_integration import (
    derivate,
    integrate,
    antiderivate,
)
from SciDataTool.Functions.sum_mean import (
    my_sum,
    my_mean,
    root_mean_square,
    root_sum_square,
)


def summing(self, values, axes_list, is_magnitude):
    """Returns the values of the field transformed or converted.
    Parameters
    ----------
    self: Data
        a Data object
    values: ndarray
        array of the field
    Returns
    -------
    values: ndarray
        values of the field
    """

    # Take magnitude before summing
    if is_magnitude:
        values = np_abs(values)

    # Apply sums, means, etc
    for axis_requested in axes_list:
        # Get axis data
        ax_val = axis_requested.values
        extension = axis_requested.extension
        index = axis_requested.index
        if axis_requested.is_pattern:
            Nper, is_aper = None, None
        else:
            Nper, is_aper = self.axes[index].get_periodicity()
        if axis_requested.name in ["time", "angle", "z"]:
            is_phys = True
            is_freqs = False
        elif axis_requested.name == "freqs":
            is_phys = False
            is_freqs = True
        else:
            is_phys = False
            is_freqs = False
        # sum over sum axes
        if extension in "sum":
            values = my_sum(values, index, Nper, is_aper)
        # root sum square over rss axes
        elif extension == "rss":
            values = root_sum_square(values, ax_val, index, Nper, is_aper, is_phys)
        # mean value over mean axes
        elif extension == "mean":
            values = my_mean(values, ax_val, index, Nper, is_aper, is_phys)
        # RMS over rms axes
        elif extension == "rms":
            values = root_mean_square(values, ax_val, index, Nper, is_aper, is_phys)
        # integration over integration axes
        elif extension == "integrate":
            values = integrate(values, ax_val, index, Nper, is_aper, is_phys)
        # integration over integration axes
        elif extension == "antiderivate":
            values = antiderivate(
                values, ax_val, index, Nper, is_aper, is_phys, is_freqs
            )
        # derivation over derivation axes
        elif extension == "derivate":
            values = derivate(values, ax_val, index, Nper, is_aper, is_phys, is_freqs)

    return values



from numpy import meshgrid, array

from SciDataTool.Functions import AxisError


def get_freqs(self):
    """Returns the frequency vector or array, if available.
    Parameters
    ----------
    self: Data
        a Data object
    Returns
    -------
    freqs : frequency vector or array
    """

    axis_names = [axis.name for axis in self.axes]

    index = None
    operation = None
    index_speed = None
    index_order = None

    if "freqs" in axis_names:
        index = axis_names.index("freqs")
    elif "frequency" in axis_names:
        index = axis_names.index("frequency")
    elif "time" in axis_names:
        index = axis_names.index("frequency")
        operation = "time_to_freqs"
    elif "speed" in axis_names and "order" in axis_names:
        index_speed = axis_names.index("speed")
        index_order = axis_names.index("order")
    else:
        raise AxisError("Cannot compute frequencies from available axes")

    if index is not None:
        freqs = self.get_axes()[index].get_values(operation=operation)
    else:  # f = speed * 60 / order
        speed = self.get_axes()[index_speed].get_values()
        order_strings = self.get_axes()[index_order].get_values().tolist()
        orders = array([int(s.split(" ")[0].replace("H", "")) for s in order_strings])
        os, so = meshgrid(orders, speed)
        freqs = so * 60 / os

    return freqs



from ...Functions.Load.import_class import import_class
from ...Classes._check import check_var
from numpy import ndarray, int32, int64, float32, float64


def _set_normalizations(self, value):
    """setter of normalizations"""
    if type(value) is dict:
        for key, obj in value.items():
            if type(obj) is dict:
                class_obj = import_class(
                    "SciDataTool.Classes", obj.get("__class__"), "normalizations"
                )
                value[key] = class_obj(init_dict=obj)
            elif isinstance(obj, (float, int, int32, int64, float32, float64)):
                Norm_ref = import_class(
                    "SciDataTool.Classes", "Norm_ref", "normalizations"
                )
                value[key] = Norm_ref(ref=obj)
            elif isinstance(obj, (ndarray, list)):
                Norm_vector = import_class(
                    "SciDataTool.Classes", "Norm_vector", "normalizations"
                )
                value[key] = Norm_vector(vector=obj)
    if type(value) is int and value == -1:
        value = dict()
    check_var("normalizations", value, "{Normalization}")
    self._normalizations = value
